#!/usr/bin/python3
"""Grafana layer module."""
import enum
import json
import re
import subprocess
import tempfile

from charmhelpers.core import unitdata
from charmhelpers.core.hookenv import (
    config,
    log,
    model_name,
    status_set,
)

from charms.layer import snap

import requests


class ChangeStatus(enum.Enum):
    """Model Snap channel states."""

    LATEST = -2
    INVALID = -1
    UPGRADE = 0
    UNCHANGED = 1
    DOWNGRADE = 2


VERSION_CHANGE_URL = "<url>"
SNAP_CODENAME_MAP = {
    "latest": -2,  # special case: we do not allow changes to the latest track
    "stable": 6,
}
TITLE_PAT = re.compile(r"\[(.*?)\] (.*)$")


def get_admin_password():
    """Get admin password from config, fall back to kv."""
    return config("admin_password") or unitdata.kv().get("grafana.admin_password")


def compute_dash_title(title, remote_app=None):
    """Compute title for dashboards."""
    if remote_app is None:
        # "local" / backward compat case -- we don't have a remote app
        return title or "Untitled"

    if not title:
        title = remote_app
    mat = TITLE_PAT.search(title)
    if mat and len(mat.groups()) == 2:
        # If the title already is in the form "[xxx] Footitle" enrich
        # the [xxx] tag with remote_app
        new_title = "[{}-{}] {}".format(mat.group(1), remote_app, mat.group(2))
    else:
        new_title = "[juju-{}] {}".format(remote_app, title)
    return new_title


def get_folders():
    """Retrieve all folders."""
    r = requests.get(
        "http://localhost:{}/api/folders".format(config("port")),
        auth=("admin", get_admin_password()),
        headers={"Accept": "application/json"},
    )
    r.raise_for_status()
    folders = r.json()
    log("get_folders: got {}".format([f["title"] for f in folders]))
    return folders


def create_folder(folder_name):
    """Create a folder through Grafana API."""
    r = requests.post(
        "http://localhost:{}/api/folders".format(config("port")),
        auth=("admin", get_admin_password()),
        headers={"Accept": "application/json"},
        data={"title": folder_name},
    )
    r.raise_for_status()
    folder = r.json()
    log("create_folder({}): created {}".format(folder_name, folder))
    return folder


def ensure_and_get_dash_folder(remote_model):
    """Get dashboard folder and create if it doesn't exist."""
    if (not remote_model) or (remote_model == model_name()):
        # Don't have a remote model, or is same as our model
        log(
            "handle_dash_folder({}): return 0 for the builtin General "
            "folder".format(remote_model)
        )
        return 0
    folders = get_folders()
    folder_name = "[juju-{}]".format(remote_model)
    matches = [f for f in folders if f["title"] == folder_name]
    if matches:
        found = matches[0]["id"]  # pick first. should only ever be one
        log("handle_dash_folder({}): found folder {}".format(remote_model, found))
        return found
    result = create_folder(folder_name)
    created = result["id"]
    log("handle_dash_folder({}): created folder {}".format(remote_model, created))
    return created


def post_dashboard(name, dashboard):
    """Upload a dashboard to Grafana."""
    headers = {"Content-Type": "application/json"}
    import_url = "http://localhost:{}/api/dashboards/db".format(config("port"))
    passwd = get_admin_password()
    if passwd is None:
        return (False, "Unable to retrieve grafana password.")
    api_auth = ("admin", passwd)
    r = requests.post(
        import_url,
        auth=api_auth,
        headers=headers,
        data=json.dumps(dashboard),
    )
    if r.status_code == 200:
        return (True, None)
    else:
        return (
            False,
            'Error importing dashboard "{}": {}'.format(name, r.text),
        )


def import_dashboard(dashboard, remote_app=None, name=None, overwrite=True):
    """Process and import a dashboard.

    Will update dashboard title if the dashboard is provided over a remote
    relation

    For dashboard provided over CMR, add the dash to a folder named after
    this model, creating this folder if necessary. Otherwise, put dashboard
    in the "General" folder

    :param dashboard: dashboard dictionary
    :param remote_app: name of remote app for dashes provided over relation
    :param remote_model: if dash is provided over CMR, name of remote model
    :param name: dash internal name
    :param overwrite: if True, overwrite existing dashboard with same id
    :return: None
    """
    if "dashboard" in dashboard:
        # bug LP1897843: dashboard json data is wrapped in {"dashboard": {<data>}}
        dashboard["dashboard"]["title"] = compute_dash_title(
            dashboard["dashboard"].get("title"), remote_app
        )

        dashboard["folderId"] = ensure_and_get_dash_folder(
            dashboard.get("source_model")
        )
        if name is None:
            name = dashboard["dashboard"].get("title") or "Untitled"
        return post_dashboard(name, dashboard)
    else:
        # correct data
        dashboard["title"] = compute_dash_title(dashboard.get("title"), remote_app)

        folder_id = ensure_and_get_dash_folder(dashboard.get("source_model"))
        data = {
            "dashboard": dashboard,
            "folderId": folder_id,
            "overwrite": overwrite,
        }
        if name is None:
            name = dashboard.get("title") or "Untitled"
        return post_dashboard(name, data)


def get_cmd_output(cmd, shell=False):
    """Get output of command passed as an argument."""
    return subprocess.check_output(cmd, shell=shell).decode("utf8")


def get_installed_package_version(name="grafana"):
    """Get current installed package version.

    Example:
        $ apt-cache policy grafana
        grafana:
          Installed: 6.5.2
          Candidate: 6.5.2
          Version table:
         *** 6.5.2 100
                100 /var/lib/dpkg/status
    Return empty str if not installed.
    """
    cmd = ["apt-cache", "policy", name]
    output = get_cmd_output(cmd)
    # output could will be empty str if package not installed.
    for line in output.splitlines():
        if "Installed:" in line:
            return line.strip().split()[-1]
    return ""


def get_deb_package_version(path):
    """Get Version field of a deb package.

    Example:
        $ dpkg -f ./grafana_6.5.2_amd64.deb Version
        6.5.2
    """
    cmd = ["dpkg", "-f", path, "Version"]
    output = get_cmd_output(cmd)
    return output.strip()


def download_file(url, proxies=None, prefix="grafana-", suffix=".deb"):
    """Download url to temp file and return path."""
    _, path = tempfile.mkstemp(prefix=prefix, suffix=suffix)
    log("downloading {} -> {}".format(url, path))
    with requests.get(url, stream=True, proxies=proxies) as r:
        r.raise_for_status()
        with open(path, "wb") as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
    return path


def parse_snap_major_version(channel):
    """Parse snap major version."""
    for codename in SNAP_CODENAME_MAP.keys():
        if channel.startswith(codename):
            return SNAP_CODENAME_MAP[codename]
    try:
        major_ver = int(channel.partition("/")[0])
    except ValueError:
        return -1
    return major_ver


def check_snap_channel(snap_name, channel):
    """Check if given channel could be configured.

    Set unit status accordingly, and return a status code from
    the ChangeStatus enum:

    LATEST: Specified "latest" track which we prohibit
    INVALID: Invalid channel
    UPGRADE: Channel configurable, needs manual upgrade
    UNCHANGED: Channel unchanged
    DOWNGRADE: Would result in downgrade of major version
    """
    installed_ver = parse_snap_major_version(snap.get_installed_channel(snap_name))
    new_ver = parse_snap_major_version(channel)
    is_refresh_available = snap.is_refresh_available(snap_name)
    if new_ver == -2:
        msg = (
            'The "latest" track is not supported; revert the channel config '
            "change to unblock"
        )
        status_set("blocked", msg)
        log(msg)
        return ChangeStatus.LATEST
    if new_ver == -1:
        msg = "Invalid snap channel: {}".format(channel)
        status_set("blocked", msg)
        log(msg)
        return ChangeStatus.INVALID
    if installed_ver == new_ver and not is_refresh_available:
        status_set("active", "Ready")
        log("Channel version unchanged: {}".format(installed_ver))
        return ChangeStatus.UNCHANGED
    if installed_ver == new_ver and is_refresh_available:
        msg = "Snap is being refreshed"
        status_set("maintenance", msg)
        log(msg)
        snap.refresh(snap_name)
        status_set("active", "Refresh complete")
        return ChangeStatus.UNCHANGED
    if installed_ver < new_ver:
        msg = (
            "PACKAGE UPGRADE REQUIRES MANUAL INTERVENTION: see {} for more info"
        ).format(VERSION_CHANGE_URL)
        status_set("blocked", msg)
        log(msg)
        return ChangeStatus.UPGRADE
    msg = (
        "PACKAGE DOWNGRADES ARE NOT SUPPORTED BY THIS CHARM - to clear "
        "this message, set config 'snap_channel' to '{}/stable'"
    ).format(installed_ver)
    status_set("blocked", msg)
    log(msg)
    return ChangeStatus.DOWNGRADE
