"""Hooks for Grafana charm.

[states]
install_packages +---------> setup_grafana +----------> do_restart
 +--> install_plugins (config.changed.install_plugins)    +
                                                          |
  +-------------------------------------------------------+
  +--> setup_backup_schedule (no grafana.backup.configured)
  +--> update_nrpe_config (nrpe-external-master.available)
  +--> check_adminuser (no grafana.admin_password.set) +----------------+
  +--> configure_sources (grafana-source, grafana.admin_password.set) <-+
  +--> import_dashboards (endpoint.dashboards.requests)

wipe_nrpe_checks (no nrpe-external-master.available)


[Helper funcs]
    * data_path(): retrieves /var/lib/grafana or /var/snap/grafana depending on
    install_method
    * check_ports(new_port): close-port (if new_port) and open-port
    * select_query(query, params=None): helper to return rows from a query
    * insert_query(query, params=None): helper to update the db
    * add_backup_api_keys(): add a new API key per org (if they don't exist)
    * validate_datasources(): validate input from Juju config (prometheus type)
    * check_datasource(ds): ds exists in database? Or insert it
    * render_custom(source, context, **parameters): JSON template to string.
    Not supported in charmhelpers, yet.
    * get_current_dashboards(port, passwd): search via Grafana API
    * get_current_dashboard_json(uid, port, passwd): also via Grafana API
    * check_and_add_dashboard(...): jsondiff between rendered dashboard
    and stored (existing and live) one
    * generate_prometheus_dashboards(gf_adminpasswd, ds): retrieves json files
    from local storage for check_and_add_dashboard to validate
    * generate_query(ds, is_default, id=None, org_id=1): to insert a new
    datasource based on grafana-source relation (with prometheus)
    * get_orgs(port, passwd): via Grafana API
    * get_main_org_id(): based on get_orgs(...) output
    * hpwgen(passwd, salt): generates a SHA256 hash for the admin password (if
    it is not shared via Juju config)

[Available states]
    main states
        grafana.installed: entrypoint (except for wipe_nrpe_checks)
        grafana.configured: apt/snap installed but unconfigured
        grafana.started: apt/snap configured and service restarted

    layer/interface or reactive framework states
        config.changed.install_plugins: tracks "install_plugins" Juju config
        nrpe-external-master.available (from interface:nrpe-external-master)
        grafana-source.available (from interface:grafana-source)
        website.available (from interface:http)
        endpoint.dashboards.requests (from interface:grafana-dashboard)

    custom (user) states
        * grafana.backup.configured: cronjob setup after Juju config param
        is set
        * grafana.admin_password.set: admin password stored after generate
        or use from Juju config

[Available hooks]
    * upgrade-charm: calls the upgrade_charm function, which removes states
    as if it was a new deployment.
    * config-changed: calls the config_changed function, which removes states
    if certain Juju config params have changed.
    * grafana-source-relation-departed: calls sources_gone (prometheus
    relation is removed)
"""
import base64
import datetime
import glob
import json
import os
import re
import shutil
import subprocess
import time

from charmhelpers import fetch
from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import (
    hookenv,
    host,
    unitdata,
)
from charmhelpers.core.templating import render

from charms.layer import snap
from charms.layer.grafana import (
    ChangeStatus,
    check_snap_channel,
    download_file,
    get_admin_password,
    get_deb_package_version,
    get_installed_package_version,
    import_dashboard,
)
from charms.reactive import (
    hook,
    remove_state,
    set_state,
    when,
    when_any,
    when_not,
)
from charms.reactive.helpers import (
    any_file_changed,
    is_state,
)

from jinja2 import Environment, FileSystemLoader, exceptions

from jsondiff import diff

import requests

import six


SVCNAME = {"snap": "snap.grafana.grafana", "apt": "grafana-server"}
SNAP_NAME = "grafana"
SNAP_DATA = "/var/snap/{}/current".format(SNAP_NAME)
SNAP_COMMON = "/var/snap/{}/common/data".format(SNAP_NAME)

GRAFANA_INI = {
    "snap": "{}/conf/grafana.ini".format(SNAP_DATA),
    "apt": "/etc/grafana/grafana.ini",
}
GRAFANA_INI_TMPL = "grafana.ini.j2"
GRAFANA_DEPS = ["libfontconfig1"]
DASHBOARDS_BACKUP_CRON = "/etc/cron.d/juju-dashboards-backup"
DASHBOARDS_BACKUP_CRON_TMPL = "juju-dashboards-backup.j2"


try:
    import sqlite3
except ImportError:
    fetch.apt_install(["python-sqlite"], fatal=True)
    import sqlite3

try:
    import pbkdf2
except ImportError:
    if six.PY3:
        fetch.apt_install(["python3-pbkdf2"], fatal=True)
    else:
        fetch.apt_install(["python-pbkdf2"], fatal=True)
    import pbkdf2

try:
    from Crypto.Hash import SHA256
except ImportError:
    if six.PY3:
        fetch.apt_install(["python3-crypto"], fatal=True)
    else:
        fetch.apt_install(["python-crypto"], fatal=True)
    from Crypto.Hash import SHA256


@when_not("grafana.installed")
def install_packages():
    """Install required packages."""
    config = hookenv.config()
    source = config.get("install_method", "apt")
    channel = config.get("snap_channel", "stable")

    kv = unitdata.kv()
    if not kv.get("install_method", False):
        kv.set("install_method", source)
    elif kv.get("install_method") != source:
        hookenv.status_set(
            "blocked",
            "Install method changes are not supported, "
            "revert install_method option to previous value",
        )
        return

    if source == "apt":
        install_opts = ("install_sources", "install_keys")
        hookenv.status_set("maintenance", "Installing grafana")

        if config.get("http_proxy"):
            proxy = {
                "http": config.get("http_proxy"),
                "https": config.get("http_proxy"),
            }
        else:
            proxy = {}

        if config.changed("install_file") and config.get("install_file"):
            hookenv.status_set(
                "maintenance",
                "Installing deb pkgs since install_file changed",
            )

            url = config.get("install_file")
            pkg_file = download_file(url, proxies=proxy)

            # if downgrade, block
            deb_version = get_deb_package_version(pkg_file)
            installed_version = get_installed_package_version()
            if deb_version < installed_version:
                message = "downgrade is not safe: {} -> {}".format(
                    installed_version, deb_version
                )
                hookenv.status_set("blocked", message)
                return

            # deb_version >= installed_version, upgrade/reinstall
            fetch.apt_install(GRAFANA_DEPS, fatal=True)
            subprocess.check_call(["dpkg", "-i", pkg_file])
        elif any(config.changed(opt) for opt in install_opts):
            hookenv.status_set("maintenance", "Installing deb pkgs")
            packages = ["grafana"]
            fetch.configure_sources(update=True)
            fetch.apt_install(packages, fatal=True)
        svcname = SVCNAME[source]
        host.service("enable", svcname)
        host.service_start(svcname)
        set_state("grafana.installed")
        hookenv.status_set("active", "Completed installing grafana")
    elif source == "snap" and (
        host.lsb_release()["DISTRIB_CODENAME"] >= "xenial"
        or host.lsb_release()["DISTRIB_CODENAME"] < "p"
    ):
        # NOTE(aluria): precise is the last supported Ubuntu release, so
        # anything below 'p' is actually newer than xenial (systemd support)
        snap.install(SNAP_NAME, channel=channel, force_dangerous=False)
        set_state("grafana.installed")
        hookenv.status_set("blocked", "Missing relations: grafana-source")
    else:
        hookenv.status_set("blocked", "Unsupported install_method")


def data_path():
    """Retrieve data store depending on install method."""
    data_dir = {"snap": SNAP_COMMON, "apt": "/var/lib/grafana"}
    source = get_install_source()
    if not source:
        remove_state("grafana.installed")

    return data_dir[source]


@when("grafana.installed")
@when("config.changed.install_plugins")
@when_not("grafana.change-block")
def install_plugins():
    """Install Grafana plugins."""
    data_dir = data_path()
    if not data_dir:
        return
    plugins_dir = "{}/plugins/".format(data_dir)

    if os.path.exists(plugins_dir):
        hookenv.log("Cleaning up currently installed plugins")
        hookenv.status_set("maintenance", "Cleaning up currently installed plugins")
        for entry in os.listdir(plugins_dir):
            entry_path = os.path.join(plugins_dir, entry)
            if os.path.isfile(entry_path):
                os.unlink(entry_path)
            elif os.path.isdir(entry_path):
                shutil.rmtree(entry_path)
        hookenv.log("Completed cleaning up grafana plugins")
        hookenv.status_set("active", "Completed cleaning up grafana plugins")

    config = hookenv.config()
    if config.get("install_plugins", False):
        hookenv.log("Installing plugins")
        hookenv.status_set("maintenance", "Installing plugins")
        plugin_file = "/tmp/grafana-plugin.tar.gz"
        for plugin_url in config.get("install_plugins").split(","):
            plugin_url = plugin_url.strip()
            hookenv.log("Downloading plugin: {!r}".format(plugin_url))
            with open(plugin_file, "wb") as f:
                r = requests.get(plugin_url, stream=True)
                for block in r.iter_content(1024):
                    f.write(block)
            hookenv.log("Extracting plugin: {!r}".format(plugin_url))
            shutil.unpack_archive(plugin_file, plugins_dir)
        hookenv.log("Completed installing grafana plugins")
        hookenv.status_set("active", "Completed installing grafana plugins")


@hook("upgrade-charm")
@when_not("grafana.change-block")
def upgrade_charm():
    """Run upgrade charm hook."""
    hookenv.status_set(
        "maintenance",
        "Forcing package update and reconfiguration on upgrade-charm",
    )
    remove_state("grafana.installed")
    remove_state("grafana.backup.configured")
    remove_state("grafana.configured")
    remove_state("grafana.nagios-setup.completed")


@hook("config-changed")
def config_changed():
    """Run config-changed hook."""
    config = hookenv.config()
    if (
        config.changed("snap_channel")
        and config["snap_channel"]
        and not config.changed("install_method")
        and config["install_method"] == "snap"
    ):
        # NOTE(aluria): install_method change (apt -> snap, and viceversa)
        # is not supported. However, snap channel change support was missing
        if snap.is_installed(SNAP_NAME):
            # Check if we should upgrade
            status = check_snap_channel(SNAP_NAME, config["snap_channel"])
            if status == ChangeStatus.UNCHANGED:
                remove_state("grafana.change-block")
            else:
                set_state("grafana.change-block")
            # We don't need to re-setup so bail out
            return
        else:
            # Install from requested channel
            snap.install(
                SNAP_NAME,
                channel=config["snap_channel"],
                force_dangerous=False,
            )
    remove_state("grafana.configured")
    if (
        config.changed("dashboards_backup_schedule")
        or config.changed("dashboards_backup_dir")
        or config.changed("port")
    ):
        remove_state("grafana.backup.configured")
    if config.changed("admin_password") or config.changed("nagios_context"):
        remove_state("grafana.admin_password.set")
    if config.changed("install_file") and config.get("install_file"):
        remove_state("grafana.installed")
    if config.changed("port"):
        # must restart grafana after port-change before updating dashboards
        remove_state("grafana.started")
    remove_state("grafana.nagios-setup.completed")


def check_ports(new_port):
    """Call to close-port/open-port when user changes the listen port.

    :param new_port: string of integer between 1-65535
    :type new_port: str
    :rtype: None
    """
    kv = unitdata.kv()
    if kv.get("grafana.port") != new_port:
        hookenv.open_port(new_port)
        if kv.get("grafana.port"):  # Dont try to close non existing ports
            hookenv.close_port(kv.get("grafana.port"))
        kv.set("grafana.port", new_port)


def select_query(query, params=None):
    """Return rows for a query.

    :param query: SQL SELECT statement
    :type query: str
    :param params: values to use in SQL statement
    :type params: list of strings
    :rtype: list of rows found
    """
    data_dir = data_path()
    if not data_dir:
        return
    conn = sqlite3.connect("{}/grafana.db".format(data_dir), timeout=30)
    cur = conn.cursor()
    if params:
        return cur.execute(query, params).fetchall()
    else:
        return cur.execute(query).fetchall()


def insert_query(query, params=None):
    """Update database.

    :param query: SQL INSERT/UPDATE/DELETE statement
    :type query: str
    :param params: values to use in SQL statement
    :type params: list of strings
    :rtype: None
    """
    data_dir = data_path()
    if not data_dir:
        return
    conn = sqlite3.connect("{}/grafana.db".format(data_dir), timeout=30)
    cur = conn.cursor()
    if params:
        cur.execute(query, params)
    else:
        cur.execute(query)
    conn.commit()


def add_backup_api_keys():
    """Generate API key per org (if it doesn't exist).

    :return: generated API key
    :rtype: str
    """
    name = "juju-dashboards-backup"
    passwd = host.pwgen(32)
    hpasswd = hpwgen(passwd, name)
    hookenv.log("Adding backup API keys for all organizations...")
    for i in select_query("SELECT id FROM org"):
        org_id = i[0]
        if select_query(
            "SELECT id FROM api_key WHERE org_id=? AND name=?",
            [org_id, name],
        ):
            hookenv.log(
                "API key {} in org {} already exists, skipping".format(name, org_id)
            )
            continue
        j = {"n": name, "k": passwd, "id": org_id}
        encoded = base64.b64encode(json.dumps(j).encode("ascii")).decode("ascii")
        stmt = (
            "INSERT INTO api_key (org_id, name, key, role, created, updated)"
            + ' VALUES (?,?,?,"Viewer",?,?)'
        )
        dtime = datetime.datetime.today().strftime("%F %T")
        params = [org_id, name, hpasswd, dtime, dtime]
        insert_query(stmt, params)

        kv = unitdata.kv()
        backup_keys = kv.get("grafana.dashboards_backup_keys", False)
        if not backup_keys:
            backup_keys = [encoded]
        else:
            backup_keys.append(encoded)
        kv.set("grafana.dashboards_backup_keys", backup_keys)

    kv = unitdata.kv()
    return kv.get("grafana.dashboards_backup_keys")


def get_install_source():
    """
    Determine if packages should be installed from apt or snap.

    :returns "snap" or "apt" based on the install_method
    :rtype Optional[str]
    """
    kv = unitdata.kv()
    source = kv.get("install_method")
    if source not in SVCNAME.keys():
        hookenv.status_set("blocked", "Unsupported install_method %s" % source)
        return None
    return source


@when("grafana.installed")
@when_not("grafana.configured")
@when_not("grafana.change-block")
def setup_grafana():
    """Install Grafana."""
    source = get_install_source()
    if not source:
        return
    grafana_ini = GRAFANA_INI[source]
    hookenv.status_set("maintenance", "Configuring grafana")
    config = hookenv.config()
    settings = {"config": config}
    smtp_auth = config.get("smtp_auth", False)
    if smtp_auth and len(smtp_auth.split(":")) == 2:
        settings["smtp_user"] = smtp_auth.split(":")[0]
        settings["smtp_password"] = smtp_auth.split(":")[1]
    render(
        source=GRAFANA_INI_TMPL,
        target=grafana_ini,
        context=settings,
        owner="root",
        group="root",
        perms=0o644,
    )
    check_ports(config.get("port"))
    set_state("grafana.configured")
    remove_state("grafana.started")
    hookenv.status_set("active", "Completed configuring grafana")


@when("grafana.started")
@when_not("grafana.backup.configured")
@when_not("grafana.change-block")
def setup_backup_shedule():
    """Copy script, create cronjob, ensure directory exists."""
    # XXX: If you add any dependencies on config items here,
    #      be sure to update config_changed() accordingly!

    config = hookenv.config()
    if config.get("dashboards_backup_schedule", False):
        hookenv.status_set("maintenance", "Configuring grafana dashboard backup")
        hookenv.log("Setting up dashboards backup job...")
        host.rsync(
            "files/dashboards_backup",
            "/usr/local/bin/dashboards_backup",
        )
        host.mkdir(config.get("dashboards_backup_dir"))
        settings = {
            "schedule": config.get("dashboards_backup_schedule"),
            "directory": config.get("dashboards_backup_dir"),
            "port": config.get("port"),
            "backup_keys": " ".join(add_backup_api_keys()),
        }
        render(
            source=DASHBOARDS_BACKUP_CRON_TMPL,
            target=DASHBOARDS_BACKUP_CRON,
            context=settings,
            owner="root",
            group="root",
            perms=0o640,
        )
        hookenv.status_set("active", "Completed configuring grafana dashboard backup")
    set_state("grafana.backup.configured")


@when("grafana.configured")
@when_not("grafana.started")
@when_not("grafana.change-block")
def restart_grafana():
    """Restart Grafana."""
    source = get_install_source()
    if not source:
        return

    svcname = SVCNAME[source]
    grafana_ini = GRAFANA_INI[source]

    if not host.service_running(svcname):
        msg = "Starting {}".format(svcname)
        hookenv.status_set("maintenance", msg)
        hookenv.log(msg)
        host.service_start(svcname)
    elif any_file_changed([grafana_ini]) or is_state("config.changed.install_plugins"):
        msg = "Restarting {}".format(svcname)
        hookenv.log(msg)
        hookenv.status_set("maintenance", msg)
        host.service_restart(svcname)
    hookenv.status_set("active", "Ready")
    set_state("grafana.started")
    hookenv.status_set("active", "Started {}".format(svcname))


@when("grafana.started")
@when_any("nrpe-external-master.available", "nrpe-external-master.connected")
@when_not("grafana.change-block")
@when_not("grafana.nagios-setup.completed")
def update_nrpe_config():
    """Update Nrpe configuration."""
    # python-dbus is used by check_upstart_job
    fetch.apt_install("python-dbus", fatal=True)
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    config = hookenv.config()

    # write the service check
    source = get_install_source()
    if not source:
        return
    svcname = SVCNAME[source]
    current_unit = nrpe.get_nagios_unit_name()
    nrpe.add_init_service_checks(nrpe_setup, [svcname], current_unit)

    # write the http check
    nrpe_setup.add_check(
        "grafana_http",
        "Grafana HTTP check",
        "check_http -I 127.0.0.1 -p {} -u /login".format(config["port"]),
    )

    nrpe_setup.write()
    set_state("grafana.nagios-setup.completed")


@when_not("nrpe-external-master.connected")
def wipe_nrpe_checks():
    """Remove Nrpe checks."""
    checks = [
        "/etc/nagios/nrpe.d/check_grafana-server.cfg",
        "/etc/nagios/nrpe.d/check_grafana.cfg",
        "/etc/nagios/nrpe.d/check_grafana_http.cfg",
        "/etc/nagios/nrpe.d/check_snap.grafana.grafana.cfg",
        "/var/lib/nagios/export/service__*_grafana-server.cfg",
        "/var/lib/nagios/export/service__*_grafana.cfg",
    ]
    for check in checks:
        for f in glob.glob(check):
            if os.path.isfile(f):
                os.unlink(f)
    remove_state("grafana.nagios-setup.completed")


@when("grafana.started")
@when("grafana-source.available")
@when("grafana.admin_password.set")
@when_not("grafana.change-block")
def configure_sources(relation):
    """Configure Grafana datasources."""
    sources = relation.datasources()
    gf_adminpasswd = get_admin_password()
    for ds in sources:
        hookenv.log("Found datasource: {}".format(str(ds)))
        # Ensure datasource is configured
        check_datasource(ds)
        if ds["type"] == "prometheus":
            generate_prometheus_dashboards(gf_adminpasswd, ds)


@hook("grafana-source-relation-departed")
def sources_gone(relation):
    """Remove Grafana datasource."""
    data_dir = data_path()
    if not data_dir:
        return
    hookenv.log("sources gone")
    hookenv.log(relation)
    # Update the datasources now with the new information that something is
    # not available.
    conv = relation.conversation()
    ds = {
        "service_name": conv.scope.split("/")[0],
        "type": conv.get_remote("type"),
        "url": conv.get_remote("url"),
        "description": conv.get_remote("description"),
    }
    hookenv.log("Removing datasource:")
    hookenv.log(ds)

    if ds:
        conn = sqlite3.connect("{}/grafana.db".format(data_dir), timeout=30)
        cur = conn.cursor()
        cur.execute(
            "DELETE FROM DATA_SOURCE WHERE  type=? AND url=?",
            (ds["type"], ds["url"]),
        )
        conn.commit()
        cur.close()
    relation.set_local = None


@when("website.available")
def configure_website(website):
    """Configure the Grafana WebUI."""
    website.configure(port=hookenv.config("port"))


def validate_datasources():
    """Verify that datasources configuration is valid, if existing.

    TODO: make sure datasources option is merged with relation data
    TODO: make sure datasources are validated
    """
    config = hookenv.config()

    if config.get("datasources", False):
        items = config["datasources"].split(",")
        if len(items) != 7:
            return False
        elif items[0] != "prometheus" and items[2] != "proxy":
            return False


def check_datasource(ds):
    """Check for and add datasources not currently in grafana DB.

    CREATE TABLE `data_source` (
    `id` INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL
    , `org_id` INTEGER NOT NULL
    , `version` INTEGER NOT NULL
    , `type` TEXT NOT NULL
    , `name` TEXT NOT NULL
    , `access` TEXT NOT NULL
    , `url` TEXT NOT NULL
    , `password` TEXT NULL
    , `user` TEXT NULL
    , `database` TEXT NULL
    , `basic_auth` INTEGER NOT NULL
    , `basic_auth_user` TEXT NULL
    , `basic_auth_password` TEXT NULL
    , `is_default` INTEGER NOT NULL
    , `json_data` TEXT NULL
    , `created` DATETIME NOT NULL
    , `updated` DATETIME NOT NULL
    , `with_credentials` INTEGER NOT NULL DEFAULT 0);
    INSERT INTO "data_source" VALUES(1,1,0,'prometheus','BootStack Prometheus','proxy','http://localhost:9090','','','',0,'','',1,'{}','2016-01-22 12:11:06','2016-01-22 12:11:11',0);
    """  # noqa E501
    # ds will be similar to:
    # {'service_name': 'prometheus',
    #  'url': 'http://10.0.3.216:9090',
    #  'description': 'Juju generated source',
    #  'type': 'prometheus',
    #  'username': 'username,
    #  'password': 'password
    # }
    data_dir = data_path()
    if not data_dir:
        return
    conn = sqlite3.connect("{}/grafana.db".format(data_dir), timeout=30)
    cur = conn.cursor()
    query = cur.execute("SELECT id, type, name, url, is_default FROM DATA_SOURCE")
    rows = query.fetchall()
    ds_name = "{} - {}".format(ds["service_name"], ds["description"])
    for row in rows:
        if row[1] == ds["type"] and row[2] == ds_name:
            hookenv.log("Datasource already exist, updating: {}".format(ds_name))
            stmt, values = generate_query(ds, row[4], id=row[0])
            print(stmt, values)
            cur.execute(stmt, values)
            conn.commit()
            conn.close()
            return
    hookenv.log("Adding new datasource: {}".format(ds_name))
    stmt, values = generate_query(ds, 0, org_id=get_main_org_id())
    cur.execute(stmt, values)
    conn.commit()
    conn.close()


# This isn't exposed in charmhelpers:
# https://github.com/juju/charm-helpers/issues/367
def render_custom(source, context, **parameters):
    """Render a template from the template folder with custom env parameters.

    source: template file name to render from
    context: template context variables
    parameters: initialization parameters for the jinja Environment

    returns the rendered template content
    """
    template_folder = os.path.join(
        hookenv.charm_dir(), "templates/dashboards/prometheus"
    )
    environment = Environment(loader=FileSystemLoader(template_folder), **parameters)
    try:
        template = environment.get_template(source)
    except exceptions.TemplateNotFound as e:
        hookenv.log(
            "Could not load template {} from {}".format(source, template_folder)
        )
        raise e
    return template.render(context)


def get_current_dashboards(port, passwd):
    """Return list of available dashboards.

    https://grafana.com/docs/grafana/latest/http_api/folder_dashboard_search/
    """
    dash_req = requests.get(
        "http://127.0.0.1:{}/api/search?type=dash-db".format(port),
        auth=("admin", passwd),
    )
    return dash_req.json() if dash_req.status_code == 200 else []


def get_current_dashboard_json(uid, port, passwd):
    """Return the dashboard revision information (fake rev if not found).

    https://grafana.com/docs/grafana/latest/http_api/dashboard/#get-dashboard-by-uid
    """  # noqa E501
    # Dashboard attributes that will be removed from the stored version (likely
    # to be different) before comparing it to the template. Value is not
    # important.
    default_dashboard = {key: 0 for key in ["version", "id", "uid"]}
    if not uid:
        return default_dashboard

    dash_req = requests.get(
        "http://127.0.0.1:{}/api/dashboards/uid/{}".format(port, uid),
        auth=("admin", passwd),
    )
    if dash_req.status_code != 200:
        return default_dashboard

    try:
        return json.loads(dash_req.text)["dashboard"]
    except json.decoder.JSONDecodeError:
        return default_dashboard


def check_and_add_dashboard(
    filename, context, prom_metrics, dash_to_uid, port, gf_adminpasswd
):
    """Verify if the rendered dashboard from template is new or do nothing.

    * Renders a template with context gathered from
    generate_prometheus_dashboards
    * Checks that metrics in the rendered dashboard have already been retrieved
    by Prometheus

    https://grafana.com/docs/grafana/latest/features/datasources/prometheus/#query-editor
    """
    dashboard_str = render_custom(
        source=filename,
        context=context,
        variable_start_string="<<",
        variable_end_string=">>",
    )
    hookenv.log("Checking Dashboard Template: {}".format(filename))
    expr = str(re.findall('"expr":(.*),', dashboard_str))
    metrics = set(re.findall("[a-zA-Z0-9]*_[a-zA-Z0-9_]*", expr))
    if not metrics:
        hookenv.log(
            "Skipping Dashboard Template: {} no metrics in template"
            " {}".format(filename, metrics)
        )
        return

    missing_metrics = set([x for x in metrics if x not in prom_metrics])
    if missing_metrics:
        hookenv.log(
            "Skipping Dashboard Template: {} missing {} metrics."
            "Missing: {}".format(
                filename,
                len(missing_metrics),
                ", ".join(missing_metrics),
            ),
            hookenv.DEBUG,
        )
        return

    dashboard_json = json.loads(dashboard_str)
    # before uploading the dashboard, we should check that it doesn't already
    # exist with same data lp#1858490
    new = dashboard_json["dashboard"]
    curr = get_current_dashboard_json(
        dash_to_uid.get(dashboard_json["dashboard"]["title"], None),
        port,
        gf_adminpasswd,
    )
    # Remove known differences between template and already stored dashboard
    # Common: 'version' (template defaults to 1)
    # Common: 'id' (template sets it null for Grafana to generate it. If the
    #         dashboard already exists in Grafana, "curr" will have an integer
    #         stored)
    # Stored dashboard: 'uid' (it is also generated by Grafana)
    del new["version"], curr["version"]
    del new["id"], curr["id"]
    del curr["uid"]
    dashboard_changed = diff(new, curr)
    if not dashboard_changed:
        hookenv.log(
            "Skipping Dashboard Template: already up to date: {}".format(filename)
        )
        return

    # Restore value. Grafana will generate a new dashboard ID
    new["id"] = None

    hookenv.log(
        "dashboard differences[{}], dashboard title[{}]".format(
            dashboard_changed,
            dashboard_json["dashboard"]["title"],
        ),
        hookenv.DEBUG,
    )

    hookenv.log("Using Dashboard Template: {}".format(filename))
    post_req = "http://127.0.0.1:{}/api/dashboards/db".format(port)
    r = requests.post(post_req, json=dashboard_json, auth=("admin", gf_adminpasswd))

    if r.status_code != 200:
        hookenv.log(
            "Posting template {} failed with error: {}".format(filename, r.text),
            hookenv.ERROR,
        )


def generate_prometheus_dashboards(gf_adminpasswd, ds):
    """Retrieve json files and validate for addition.

    :param gf_adminpasswd: superuser password
    :type gf_adminpasswd: str
    :param ds: datasource attributes (name, url, ...)
    :type ds: list of strings
    :rtype: None
    """
    # prometheus_host = ds
    ds_name = "{} - {}".format(ds["service_name"], ds["description"])
    config = hookenv.config()

    # https://prometheus.io/docs/prometheus/latest/querying/api/#querying-label-values
    try:
        response = requests.get("{}/api/v1/label/__name__/values".format(ds["url"]))
    except requests.exceptions.ConnectionError as e:
        hookenv.log(
            "Exception while trying to reach prometheus API: {}".format(e),
            "ERROR",
        )
        hookenv.status_set(
            "blocked", "Exception reaching prometheus API whilst updating dashboards"
        )
        return

    if response.status_code != 200:
        hookenv.log(
            "Could not reach prometheus API status code: "
            " {}".format(response.status_code),
            "ERROR",
        )
        return

    current_dashboards = get_current_dashboards(config["port"], gf_adminpasswd)
    dash_to_uid = {
        dash["title"]: dash["uid"]
        for dash in current_dashboards
        if "title" in dash and "uid" in dash
    }

    prom_metrics = response.json()["data"]
    templates_dir = "templates/dashboards/prometheus"
    context = {
        "datasource": ds_name,
        "external_network": config["external_network"],
        "bcache_enabled": "bcache_cache_hit_ratio" in prom_metrics,
        "bonds_enabled": "bond_status" in prom_metrics,
        "conntrack_enabled": "nf_conntrack_max" in prom_metrics,
    }
    # we ignore some phrases picked up by the regex because they're not real
    # metric names.  Let's add them to the list of metrics to make life easy.
    ignore_metrics = [
        "by_cloud_status",
        "group_left",
        "count_values",
        "nova_service_status",
        "hypervisor_hostname",
        "external_network",
        "ip_type",
        "instance_state",
        "network_name",
        "subnet_name",
        "ip_status",
        "neutron_net",
        config["external_network"],
        "neutron_public_ip_usage",
    ]
    prom_metrics.extend(ignore_metrics)

    for filename in os.listdir(templates_dir):
        check_and_add_dashboard(
            filename,
            context,
            prom_metrics,
            dash_to_uid,
            config["port"],
            gf_adminpasswd,
        )


def generate_query(ds, is_default, id=None, org_id=1):
    """Generate SQL query to add or update a datasource.

    :param ds: datasource attributes (name, url...)
    :type ds: list
    :param is_default: is it the default datasource
    :type is_default: bool
    :param id: use a datasource ID (ie. to update it)
    :type id: int
    :param org_id: which Grafana organization to use (defaults to main)
    :type org_id: int
    :return: SQL statement and params
    :rtype: tuple
    """
    if not id:
        fields = [
            "org_id",
            "version",
            "type",
            "name",
            "access",
            "url",
            "is_default",
            "created",
            "updated",
        ]

        dtime = datetime.datetime.today().strftime("%F %T")
        values = [
            org_id,
            0,
            ds["type"],
            "{} - {}".format(ds["service_name"], ds["description"]),
            "proxy",
            ds["url"],
            is_default,
            dtime,
            dtime,
        ]

        config = hookenv.config()
        if config["install_method"] == "apt" or config["snap_channel"].startswith("7/"):
            fields.append("uid")
            # Grafana's SQLite3 database has a unique constraint on the org_id
            # and uid field, so generate one based on a hash of the data source
            # url which is unique.
            # The 32 bit mask is necessary because Python's 64 bit ints are too
            # large for SQLite3.
            values.append(hash(ds["url"]) & 0xFFFFFFFF)
            # latest grafana will fail to accept the datasource
            # unless these fields are default as follows
            fields.append("json_data")
            values.append("{}")
            fields.append("secure_json_data")
            values.append("{}")
            fields.append("read_only")
            values.append(0)

        # Used in connecting a mysql source
        if "database" in ds:
            hookenv.log("Adding fields for the database")
            fields.append("database")
            values.append(ds["database"])

        if ds["type"] == "mysql":
            # There's only one username/password on the interface to we coopt
            # it for the mysql login in the mysql case
            fields.append("user")
            values.append(ds["username"])
            fields.append("password")
            values.append(ds["password"])

        if "username" in ds and "password" in ds and ds["type"] != "msyql":
            # We use the username and password in a backward compatible way
            # for basic_auth when it's not a mysql data source.
            hookenv.log("Adding basic_auth info")
            fields.append("basic_auth")
            fields.append("basic_auth_user")
            fields.append("basic_auth_password")
            values.append(1)
            values.append(ds["username"])
            values.append(ds["password"])
        else:
            fields.append("basic_auth")
            values.append(0)

        values_placeholders = "?, " * len(values)
        # Remove the space [, ] at the end of the generated string.
        values_placeholders = values_placeholders[:-2]

        stmt = """
         INSERT INTO DATA_SOURCE ({})
         VALUES ({})
        """.format(
            ", ".join(fields), values_placeholders
        )

        hookenv.log("Statement is: \n{}".format(stmt))
        hookenv.log(values)

    else:
        if "username" in ds and "password" in ds:
            stmt = (
                "UPDATE DATA_SOURCE SET basic_auth_user = ?, "
                "basic_auth_password = ?, basic_auth = 1"
            )
            values = (ds["username"], ds["password"])
        else:
            stmt = (
                "UPDATE DATA_SOURCE SET basic_auth_user = ?, "
                "basic_auth_password = ?, basic_auth = 0"
            )
            values = ("", "")
    return (stmt, values)


def get_orgs(port, passwd):
    """Return list of available orgs.

    https://grafana.com/docs/grafana/latest/http_api/org/
    """
    req = requests.get(
        "http://127.0.0.1:{}/api/orgs".format(port),
        auth=("admin", passwd),
    )
    return req.json() if req.status_code == 200 else []


def get_main_org_id():
    """Get Main organization ID."""
    config = hookenv.config()
    port = config["port"]
    admin_password = get_admin_password()
    orgs = get_orgs(port, admin_password)
    return orgs[0]["id"] if orgs else 1


@when("grafana.started")
@when_not("grafana.admin_password.set")
@when_not("grafana.change-block")
def check_adminuser():
    """Create Adminuser if not existing.

    CREATE TABLE `user` (
    `id` INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL
    , `version` INTEGER NOT NULL
    , `login` TEXT NOT NULL
    , `email` TEXT NOT NULL
    , `name` TEXT NULL
    , `password` TEXT NULL
    , `salt` TEXT NULL
    , `rands` TEXT NULL
    , `company` TEXT NULL
    , `org_id` INTEGER NOT NULL
    , `is_admin` INTEGER NOT NULL
    , `email_verified` INTEGER NULL
    , `theme` TEXT NULL
    , `created` DATETIME NOT NULL
    , `updated` DATETIME NOT NULL
    );
    INSERT INTO "user" VALUES(1,0,'admin','root+bootstack-ps45@canonical.com','BootStack Team','309bc4e78bc60d02dc0371d9e9fa6bf9a809d5dc25c745b9e3f85c3ed49c6feccd4ffc96d1db922f4297663a209e93f7f2b6','LZeJ3nSdrC','hseJcLcnPN','',1,1,0,'light','2016-01-22 12:00:08','2016-01-22 12:02:13');
    """  # noqa E501
    # XXX: If you add any dependencies on config items here,
    #      be sure to update config_changed() accordingly!

    config = hookenv.config()
    passwd = config.get("admin_password", False)
    if not passwd:
        passwd = host.pwgen(16)
        kv = unitdata.kv()
        kv.set("grafana.admin_password", passwd)

    try:
        stmt = "UPDATE user SET email=?, name='BootStack Team'"
        stmt += ", password=?, theme='light'"
        stmt += " WHERE id = ?"

        data_dir = data_path()
        if not data_dir:
            return
        conn = sqlite3.connect("{}/grafana.db".format(data_dir), timeout=30)
        cur = conn.cursor()
        # Check if we have more than one user in the db
        for x in range(0, 5):
            hookenv.log("Checking for existing users, loop %d of 6" % (x + 1))
            time.sleep(5)
            results = cur.execute("SELECT count(login) from user")
            usercount = results.fetchall()[0][0]
            if usercount > 0:
                break
        else:
            hookenv.log("We didn't find any users!")
            return

        query = cur.execute("SELECT id, login, salt FROM user")
        for row in query.fetchall():
            if row[1] == "admin":
                nagios_context = config.get("nagios_context", "UNKNOWN")
                email = "root+%s@canonical.com" % nagios_context
                hpasswd = hpwgen(passwd, row[2])
                if hpasswd:
                    cur.execute(stmt, (email, hpasswd, row[0]))
                    conn.commit()
                    hookenv.log("[*] admin password updated on database")
                else:
                    hookenv.log("Could not update user table: hpwgen func failed")
                break
        conn.close()
    except sqlite3.OperationalError as e:
        hookenv.log("check_adminuser::sqlite3.OperationError: {}".format(e))
        return
    set_state("grafana.admin_password.set")


def hpwgen(passwd, salt):
    """Generate a random password."""
    hpasswd = pbkdf2.PBKDF2(passwd, salt, 10000, SHA256).hexread(50)
    return hpasswd


def ensure_dash_deserialized(dash):
    """Deserialize dash.

    Per the dashboard interface the dashboards transmitted over a relation
    should already be de-serialized json, but older dashboard providers used
    to send a json string, so attempt to deserialize if we find a string
    """
    if isinstance(dash, str):
        dash = json.loads(dash)
    return dash


@hook("dashboards-relation-joined")
def dashboards_joined(dashboards):
    """Process dashboards from related charms."""
    hookenv.log("hook %s triggered" % hookenv.hook_name(), "INFO")

    remote_unit = os.environ.get("JUJU_REMOTE_UNIT", None)
    if not remote_unit:
        hookenv.log("remote_unit is None, return", "WARNING")
        return

    dashboard_str = hookenv.relation_get("dashboard", remote_unit)  # json str
    if not dashboard_str:
        hookenv.log("dashboard field is empty, return", "WARNING")
        return

    dashboard = ensure_dash_deserialized(dashboard_str)
    name = hookenv.relation_get("name", remote_unit)  # dashboard name
    remote_app = hookenv.remote_service_name()

    success, reason = import_dashboard(dashboard, remote_app=remote_app, name=name)
    if success:
        hookenv.log("import_dashboard {} success".format(name), "INFO")
    else:
        hookenv.log("import_dashboard {} failed: {}".format(name, reason), "ERROR")


@when("grafana.started", "endpoint.dashboards.has_requests")
@when_not("grafana.change-block")
def import_dashboards(dashboards):
    """Import available dashboards."""
    kv = unitdata.kv()
    dashboard_digests = kv.get("dash_digests", {})
    remote_app = hookenv.remote_service_name()
    new_requests = set([r.request_id for r in dashboards.new_requests])
    for req in dashboards.all_requests:
        dash = ensure_dash_deserialized(req.dashboard)
        is_new = req.request_id in new_requests
        digest = dash.get("digest")
        name = req.name
        hookenv.log(
            "import_dashboards: {}, digest {}, is_new: {}".format(name, digest, is_new)
        )
        if not digest and not is_new:
            # Backward compat: if we don't see a digest in the payload, it's
            # coming from an older provider. Skip unless it's new
            continue
        if dashboard_digests.get(name) != digest or is_new:
            success, reason = import_dashboard(dash, remote_app, name)
            dashboard_digests[name] = digest
            hookenv.log(
                "import_dashboards {}, digest {}: success: {} [{}]".format(
                    name, digest, success, reason
                )
            )
            if is_new:
                req.respond(success, reason)

    kv.set("dash_digests", dashboard_digests)
