"""Encapsulate prometheus-openstack-exporter testing."""
import json
import logging
import time
import unittest

import requests

import zaza.model as model


TEST_TIMEOUT = 600
DEFAULT_API_PORT = "3000"
DEFAULT_API_URL = "/"
DEFAULT_BACKUP_DIRECTORY = "/srv/backups"


class BaseGrafanaTest(unittest.TestCase):
    """Base for Prometheus-openstack-exporter charm tests."""

    _admin_pass = None

    @classmethod
    def setUpClass(cls):
        """Set up tests."""
        super(BaseGrafanaTest, cls).setUpClass()
        cls.model_name = model.get_juju_model()
        cls.application_name = "grafana"
        cls.lead_unit_name = model.get_lead_unit_name(
            cls.application_name, model_name=cls.model_name
        )
        cls.units = model.get_units(cls.application_name, model_name=cls.model_name)
        cls.grafana_ip = model.get_app_ips(cls.application_name)[0]

    def get_unit_status(self, unit):
        """Get grafana unit workload status."""
        u = model.get_unit_from_name(unit)
        return u.workload_status_message

    @property
    def admin_password(self):
        """Get grafana admin password."""
        if self._admin_pass is None:
            action = model.run_action(
                self.lead_unit_name,
                "get-admin-password",
                raise_on_failure=True,
            )
            self._admin_pass = action.data["results"]["password"]
        return self._admin_pass

    def get_and_check_dashboards(self, cond):
        """Retrieve dashboard from grafana.

        Check if cond(dashboards) is true, otherwise retry up to TEST_TIMEOUT
        """
        timeout = time.time() + TEST_TIMEOUT
        dashboards = []
        while True:
            req = requests.get(
                "http://{host}:{port}"
                "/api/search?dashboardIds".format(
                    host=self.grafana_ip, port=DEFAULT_API_PORT
                ),
                auth=("admin", self.admin_password),
            )
            self.assertTrue(req.status_code == 200)
            dashboards = json.loads(req.text)
            if time.time() > timeout or cond(dashboards):
                break
            time.sleep(30)
        return dashboards

    def remote_sed(self, unit, file, expr):
        """Run sed on remote unit."""
        cmd = ["sed", "-i", "-e", expr, file]
        model.run_on_unit(unit, " ".join(cmd))

    def verify_iterative_backups(self):
        """Wait for new backups to appear to ensure they work."""
        time.sleep(65)
        backupdir = model.run_on_unit(
            self.lead_unit_name,
            "ls {directory}".format(directory=DEFAULT_BACKUP_DIRECTORY),
        )
        self.assertTrue("No such file or directory" not in backupdir["Stderr"])
        backups_start = model.run_on_unit(
            self.lead_unit_name,
            "ls {project}/{backup}".format(
                project=DEFAULT_BACKUP_DIRECTORY, backup=backupdir["Stdout"].strip()
            ),
        )
        self.assertTrue(".json" in backups_start["Stdout"])
        number_of_backups_start = backups_start["Stdout"].count(".json")
        self.assertTrue(number_of_backups_start >= 1)
        time.sleep(65)
        backups_end = model.run_on_unit(
            self.lead_unit_name,
            "ls {project}/{backup}".format(
                project=DEFAULT_BACKUP_DIRECTORY, backup=backupdir["Stdout"].strip()
            ),
        )
        number_of_backups_end = backups_end["Stdout"].count(".json")
        self.assertTrue(number_of_backups_end > number_of_backups_start)


class CharmOperationTest(BaseGrafanaTest):
    """Verify operations."""

    def test_01_api_ready(self):
        """Verify if the API is ready.

        Curl the api endpoint.
        We'll retry until the TEST_TIMEOUT.
        """
        test_command = "{} -I 127.0.0.1 -p {} -u {}".format(
            "/usr/lib/nagios/plugins/check_http",
            DEFAULT_API_PORT,
            DEFAULT_API_URL,
        )
        timeout = time.time() + TEST_TIMEOUT
        while time.time() < timeout:
            response = model.run_on_unit(self.lead_unit_name, test_command)
            if response["Code"] == "0":
                return
            logging.warning(
                "Unexpected check_http response: {}. Retrying in 30s.".format(response)
            )
            time.sleep(30)

        # we didn't get rc=0 in the allowed time, fail the test
        self.fail(
            "http port didn't respond to the command \n"
            "'{test_command}' as expected.\n"
            "Result: {result}".format(test_command=test_command, result=response)
        )

    def test_02_nrpe_http_check(self):
        """Verify nrpe check exists."""
        logging.debug(
            "Verify the nrpe check is created and has the required content..."
        )
        expected_nrpe_check = (
            "command[check_grafana_http]=/usr/lib/nagios/plugins/check_http "
            "-I 127.0.0.1 -p 3000 -u /login"
        )
        cmd = "cat /etc/nagios/nrpe.d/check_grafana_http.cfg"
        result = model.run_on_unit(self.lead_unit_name, cmd)
        code = result.get("Code")
        if code != "0":
            raise model.CommandRunFailed(cmd, result)
        content = result.get("Stdout")
        self.assertTrue(expected_nrpe_check in content)

    def test_10_grafana_imported_dashboards(self):
        """Check that Grafana dashboards expected are there."""
        dashboards = self.get_and_check_dashboards(lambda dashes: len(dashes) >= 4)
        self.assertTrue(len(dashboards) >= 4)
        self.assertTrue(
            all(
                dash.get("type", "") == "dash-db"
                and dash.get("uri", "").startswith("db/juju-")
                for dash in dashboards
            )
        )

    def test_11_dashboard_upgrade(self):
        """Test upgraded dashboards are imported."""
        exporter = "prometheus-libvirt-exporter/0"
        self.remote_sed(
            exporter,
            "/var/lib/juju/agents/unit-prometheus-libvirt-exporter-0/charm/"
            "files/grafana-dashboards/libvirt.json",
            "s/Libvirt/FunctionalTestMarker/",
        )
        model.run_on_unit(
            exporter, "JUJU_HOOK_NAME=upgrade-charm ./hooks/upgrade-charm"
        )
        model.block_until_all_units_idle()

        def chk(dashes):
            marker_dashes = [d for d in dashes if "FunctionalTestMarker" in d["title"]]
            return len(dashes) >= 4 and len(marker_dashes) == 1

        dashes = self.get_and_check_dashboards(chk)
        self.assertTrue(chk(dashes))

        self.remote_sed(
            exporter,
            "/var/lib/juju/agents/unit-prometheus-libvirt-exporter-0/charm/"
            "files/grafana-dashboards/libvirt.json",
            "s/FunctionalTestMarker/Libvirt/",
        )
        model.run_on_unit(
            exporter, "JUJU_HOOK_NAME=upgrade-charm ./hooks/upgrade-charm"
        )
        model.block_until_all_units_idle()

    def test_12_grafana_create_user(self):
        """Run the create-user action."""
        params = {
            "name": "FooUser",
            "login": "foouser",
            "email": "foo@example.com",
            "password": "sikkrit",
            "role": "Viewer",
        }
        action = model.run_action(
            self.lead_unit_name,
            "create-user",
            action_params=params,
        )
        self.assertTrue(action.data["results"]["Code"] == "0")
        time.sleep(30)  # Dirty hack to overcome race condition
        req = requests.get(
            "http://{host}:{port}/api/org/".format(
                host=self.grafana_ip, port=DEFAULT_API_PORT
            ),
            auth=("foouser", "sikkrit"),
        )
        self.assertTrue(req.status_code == 200)
        self.assertTrue("name" in req.json())

    def test_13_grafana_dashboard_backup(self):
        """Run dashboard-backup items."""
        model.set_application_config(
            self.application_name,
            {"dashboards_backup_schedule": "* * * * *"},
        )
        model.block_until_all_units_idle()
        crontab = model.run_on_unit(
            self.lead_unit_name,
            "cat /etc/cron.d/juju-dashboards-backup",
        )
        self.assertTrue(DEFAULT_BACKUP_DIRECTORY in crontab["Stdout"])
        self.assertTrue(DEFAULT_API_PORT in crontab["Stdout"])
        self.verify_iterative_backups()

    def test_14_grafana_dashboard_backup_port_change(self):
        """Run dashboard backup and test a port change."""
        port = "4321"
        model.set_application_config(
            self.application_name, {"dashboards_backup_schedule": "* * * * *"}
        )
        model.block_until_all_units_idle()
        model.set_application_config(self.application_name, {"port": port})
        model.block_until_all_units_idle()
        crontab = model.run_on_unit(
            self.lead_unit_name, "cat /etc/cron.d/juju-dashboards-backup"
        )
        self.assertTrue(port in crontab["Stdout"])
        self.verify_iterative_backups()

    def test_15_grafana_datasource_updates(self):
        """Change the port on the associated datasource.

        Ensures the model returns to idle after changing the prometheus port post
        deployment.

        As per LP #1893320
        """
        datasource_application = "prometheus"
        port_config = {"web-listen-port": "1234"}
        model.set_application_config(datasource_application, port_config)
        model.block_until_all_units_idle()


class SnappedGrafanaTest(BaseGrafanaTest):
    """Verify Grafana installed as a snap."""

    def test_50_snap_upgrade(self):
        """Test a snap upgrade from 6 to 7."""
        model.set_application_config(
            self.application_name, {"snap_channel": "7/stable"}
        )
        model.block_until_unit_wl_status(self.lead_unit_name, "blocked")
        model.block_until_all_units_idle()
        status_message = self.get_unit_status(self.lead_unit_name)
        self.assertTrue(
            "PACKAGE UPGRADE REQUIRES MANUAL INTERVENTION" in status_message
        )
        action = model.run_action(self.lead_unit_name, "do-upgrade")
        self.assertEqual(action.data["status"], "completed")
        self.assertTrue("Upgraded" in action.data["results"]["result"])

    def test_51_snap_downgrade_block(self):
        """Test that after the previous upgrade, downgrades are blocked."""
        model.set_application_config(
            self.application_name, {"snap_channel": "6/stable"}
        )
        model.block_until_unit_wl_status(self.lead_unit_name, "blocked")
        model.block_until_all_units_idle()
        status_message = self.get_unit_status(self.lead_unit_name)
        self.assertTrue("PACKAGE DOWNGRADES ARE NOT SUPPORTED" in status_message)
        model.set_application_config(  # unblock unit again
            self.application_name, {"snap_channel": "7/stable"}
        )
        model.block_until_all_units_idle()
        downgrade_cmd = "sudo snap refresh grafana --channel=6/stable"
        model.run_on_unit(self.lead_unit_name, downgrade_cmd)
        model.set_application_config(  # set it back to 6
            self.application_name, {"snap_channel": "6/stable"}
        )
        model.block_until_all_units_idle()

    def test_52_latest(self):
        """Test that track 'latest' is result in blocked status."""
        model.set_application_config(
            self.application_name, {"snap_channel": "latest/stable"}
        )
        model.block_until_unit_wl_status(self.lead_unit_name, "blocked")
        model.block_until_all_units_idle()
        status_message = self.get_unit_status(self.lead_unit_name)
        self.assertTrue(
            'The "latest" track is not supported; revert ' in status_message
        )
        model.set_application_config(  # unblock unit again
            self.application_name, {"snap_channel": "6/stable"}
        )
        model.block_until_all_units_idle()


class AptGrafanaTest(BaseGrafanaTest):
    """Verify Grafana installed from apt repos."""

    def test_50_apt_upgrade(self):
        """Test an apt upgrade."""
        action = model.run_action(self.lead_unit_name, "do-upgrade")
        self.assertEqual(action.data["status"], "completed")
        self.assertTrue("Upgraded" in action.data["results"]["result"])
