"""Unit tests module."""
import sys
import unittest
from os.path import isfile
from unittest import mock
from unittest.mock import call

from charmhelpers.core import unitdata

sys.modules["charms.layer.snap"] = mock.Mock()


from charms.layer import grafana  # noqa E402
from charms.layer.grafana import (  # noqa E402
    compute_dash_title,
    download_file,
    get_cmd_output,
    get_deb_package_version,
    get_installed_package_version,
)
from reactive import grafana as grafana_reactive  # noqa E402


class GrafanaTestCase(unittest.TestCase):
    """Unit tests class for Grafana."""

    def test_get_installed_package_version(self):
        """Test get_installed_package_version with basic packages."""
        # these packages are installed by layer:basic
        packages = [
            "python3-pip",
            "python3-setuptools",
            "python3-yaml",
            "python3-dev",
            "python3-wheel",
            "build-essential",
        ]
        for package in packages:
            # get the version with a different cmd
            cmd = "apt list --installed | grep {}".format(package)
            apt_list = get_cmd_output(cmd, shell=True)
            version = get_installed_package_version(package)
            self.assertIn(package, apt_list)
            self.assertIn(version, apt_list)

    def test_download_file_and_deb_package_version(self):
        """Test download_file and get_deb_package_version with hello package."""
        url = (
            "http://http.us.debian.org/debian/pool"
            "/main/h/hello/hello_2.10-2_amd64.deb"
        )
        path = download_file(url, prefix="hello-")
        self.assertTrue(isfile(path))
        version = get_deb_package_version(path)
        self.assertEqual(version, "2.10-2")

    @mock.patch("charmhelpers.fetch.apt_install")
    @mock.patch("charmhelpers.contrib.charmsupport.nrpe.NRPE")
    @mock.patch("charmhelpers.contrib.charmsupport.nrpe.get_nagios_hostname")
    @mock.patch("charmhelpers.contrib.charmsupport.nrpe.get_nagios_unit_name")
    @mock.patch("charmhelpers.core.hookenv.config")
    def test_update_nrpe_config(
        self,
        mock_hookenv_config,
        mock_get_nagios_unit_name,
        mock_get_nagios_hostname,
        mock_nrpe,
        *args
    ):
        """Test nrpe configuration update."""
        sys.modules["charms.layer"] = mock.MagicMock()
        sys.modules["charms.layer.grafana"] = mock.MagicMock()
        mock_hookenv_config.return_value = {"port": 1234}
        mock_get_nagios_hostname.return_value = "testunit"
        mock_get_nagios_unit_name.return_value = "my_grafana/0"
        unitdata.kv().update({"install_method": "snap"}, prefix="")
        grafana_reactive.update_nrpe_config()
        calls = [
            call(hostname="testunit"),
            call().add_check(
                shortname="snap.grafana.grafana",
                description="process check {my_grafana/0}",
                check_cmd="check_systemd.py snap.grafana.grafana",
            ),
            call().add_check(
                "grafana_http",
                "Grafana HTTP check",
                "check_http -I 127.0.0.1 -p 1234 -u /login",
            ),
            call().write(),
        ]
        mock_nrpe.assert_has_calls(calls)

    def test_compute_dashboard_title(self):
        """Test mangling of dashboard title."""
        inputs_expect = (
            (("foo", "app"), "[juju-app] foo"),
            (("[x] foo", "app"), "[x-app] foo"),
            ((None, "app"), "[juju-app] app"),
            (("bar", None), "bar"),
            ((None, None), "Untitled"),
        )
        for inputs, expect in inputs_expect:
            self.assertEqual(compute_dash_title(*inputs), expect)

    @mock.patch.object(grafana, "model_name")
    @mock.patch.object(grafana, "get_folders", return_value=[])
    @mock.patch.object(grafana, "create_folder", return_value={"id": 77})
    def test_ensure_and_get_dash_folder_no_folder(
        self, mock_create, _mock_get, _mock_model_name
    ):
        """Test if a folder gets created for a CMR."""
        folder = grafana.ensure_and_get_dash_folder("foo-model")
        self.assertEqual(folder, 77)
        mock_create.assert_called_once_with("[juju-foo-model]")

    @mock.patch.object(grafana, "model_name", return_value="foo-model")
    @mock.patch.object(grafana, "create_folder")
    def test_ensure_and_get_dash_folder_same_model(self, mock_create, _mock_model_name):
        """Test: no folder created if provider is in same model."""
        folder = grafana.ensure_and_get_dash_folder("foo-model")
        self.assertEqual(folder, 0)
        mock_create.assert_not_called()

    @mock.patch.object(grafana, "model_name")
    @mock.patch.object(
        grafana,
        "get_folders",
        return_value=[{"id": 7, "title": "[juju-foo-model]"}],
    )
    @mock.patch.object(grafana, "create_folder")
    def test_ensure_and_get_dash_folder_folder_exists(
        self, mock_create, _mock_get, _mock_model_name
    ):
        """Test: no folder created if provider is in same model."""
        folder = grafana.ensure_and_get_dash_folder("foo-model")
        self.assertEqual(folder, 7)
        mock_create.assert_not_called()

    @mock.patch.object(grafana, "model_name", return_value="foo-model")
    @mock.patch.object(grafana, "post_dashboard")
    def test_import_dashboard(self, mock_post, _mock_model_name):
        """Test importing dashboards."""
        test_dash = {
            "dashboard": {"title": "test-title"},
        }
        grafana.import_dashboard(test_dash, remote_app="foo-app", name="dash-name")
        mock_post.assert_called_once_with(
            "dash-name",
            {"dashboard": {"title": "[juju-foo-app] test-title"}, "folderId": 0},
        )
